from typing import List, Tuple
from .inference import Filter
from .inference import Engine


class Evaluator:
    """
    Evaluate a regex against the ground truth patterns
    """
    @staticmethod
    def evaluate(
            regex: str, patterns: List[str]) -> Tuple[float, float, float]:
        """
        Args:
            - regex: regex to be evaluated
            - patterns: patterns to be matched by the regex
        Returns:
            - precision: describe how well the regex describe the patterns
            - recall: describe how well the regex captures the patterns
            - f1: combined score for precision and recall
        """
        return Evaluator.evaluate_regex_list([regex], patterns)

    @staticmethod
    def evaluate_regex_list(
            regex_list: List[str], patterns: List[str]) -> Tuple[float, float, float]:
        """
        Args:
            - regex_list: regex to be evaluated
            - patterns: patterns to be matched by the regex_list
        Returns:
            - precision: describe how well each regex in the regex list describe the patterns
            - recall: describe how well the entire regex list match the patterns
            - f1: combined score for precision and recall
        """

        recall = Evaluator.recall(regex_list, patterns)
        precision = Evaluator.precision(regex_list, patterns)
        if recall == 0. or precision == 0.:
            f1 = 0.
        else:
            f1 = 2. / (1. / precision + 1. / recall)
        return precision, recall, f1

    @staticmethod
    def precision(regex_list: List[str], patterns: List[str]) -> float:
        divided_patterns = Engine.divide_patterns(regex_list, patterns)
        precisions = []
        for i in range(len(divided_patterns)):
            negative_patterns = Evaluator._collect_negative_patterns(
                i, divided_patterns)
            precision = Evaluator._regex_precision(
                regex_list[i], divided_patterns[i], negative_patterns)
            precisions.append(precision)
        precision = sum(precisions) / len(precisions)
        return precision

    @staticmethod
    def recall(regex_list: List[str], patterns: List[str]) -> float:
        """
        Recall evaluate how well the regex capture the patterns presented.

        Args:
            - regex: whole regex consists of multiple sub-regex
            - patterns: the patterns in the future or not presented but should be captured by the regex.
        """
        regex = Engine.merge_regex_sequence(regex_list)
        return len(Filter.match(regex, patterns)) / len(patterns)

    @staticmethod
    def _collect_negative_patterns(
            target_regex_index: int, divided_patterns: List[List[str]]) -> List[str]:
        negative_patterns = []
        for not_i in [j for j in range(
                len(divided_patterns)) if j != target_regex_index]:
            negative_patterns.extend(divided_patterns[not_i])
        return negative_patterns

    @staticmethod
    def _regex_precision(
            sub_regex: str, positive_patterns: List[str], negative_patterns: List[str]) -> float:
        """
        Precision evaluate how precise or explainable is the regex on the target patterns.

        Because my goal is that each sub-regex should exactly match its target patterns,
        the positive patterns and negative patterns for the sub-regex is defined as follows:

        * positive_patterns: pattern presented previously and matched by the sub-regex
        * negative_patterns: pattern not hosted by the sub-regex.
        """
        if positive_patterns:
            return len(Filter.match(sub_regex, positive_patterns)) / \
                len(Filter.match(sub_regex,
                                 positive_patterns + negative_patterns))
        else:
            return 0.0



from typing import List, Tuple
from threading import Thread
# from multiprocessing import Process as Thread
from multiprocessing import Queue
from more_itertools import chunked
import random
from .engine import Engine
from .fado import FAdoEngine, FAdoAIEngine
from ..evaluator import Evaluator

__all__ = ['Inference']


class Candidate(Thread):
    def __init__(self, engine, train_patterns, val_patterns, queue):
        Thread.__init__(self)
        self.value = None
        self._engine = engine
        self._train_patterns = train_patterns
        self._val_patterns = val_patterns
        self._q = queue

    def run(self):
        regex_list = self._engine.get_regex_sequence(
            self._train_patterns)
        _, _, f1 = Evaluator.evaluate_regex_list(
            regex_list, self._val_patterns)
        self.value = (f1, regex_list)
        self._q.put((f1, regex_list))

    @staticmethod
    def get_best(candidates: List['Candidate']) -> str:
        for worker in candidates:
            worker.start()
        for worker in candidates:
            worker.join()
        regex_list = [Candidate.get_value(worker) for worker in candidates]
        sorted_results = sorted(regex_list, key=lambda x: x[0], reverse=True)
        return Engine.merge_regex_sequence(sorted_results[0][1])

    @staticmethod
    def get_value(candidate) -> Tuple[float, List[str]]:
        value = candidate.value
        if value is None:
            value = candidate._q.get()
        return value


class Inference:
    def __init__(self, *args, **kwargs):
        if 'engine' in kwargs:
            if kwargs['engine'] == 'fado+ai':
                del kwargs['engine']
                self._engine = FAdoAIEngine(*args, **kwargs)
            elif kwargs['engine'] == 'ai':
                del kwargs['engine']
                self._engine = Engine(*args, **kwargs)
            elif kwargs['engine'] == 'fado':
                del kwargs['engine']
                self._engine = FAdoEngine(*args, **kwargs)
        else:
            self._engine = FAdoAIEngine(*args, **kwargs)
        if 'verbose' in kwargs:
            self._verbose = kwargs['verbose']
        else:
            self._verbose = False

    def run(self, train_patterns: List[str],
            val_patterns: List[str] = [], n_fold: int = 10, train_rate: float = 1.) -> str:
        """
        Args:
            - train_patterns: The patterns to be infered from.
            - val_patterns: The validation patterns for selecting the best regex.
                If not provided, validation data would be randomly and non-repetively selected from training data.
            - n_fold: repeating validation count.
            - train_rate: Ratio of training data selected for each fold (used only when val_patterns is not provided.)
        Return:
            - regex: The infered regex
        """
        if val_patterns:
            regex = self._infer_with_fix_val_patterns(
                train_patterns, val_patterns, n_fold=n_fold)
        else:
            regex = self._infer_with_cross_val_patterns(
                train_patterns, n_fold=n_fold, total_train_rate=train_rate * n_fold)
        return regex

    def _infer_with_fix_val_patterns(
            self, train_patterns: List[str], val_patterns: List[str], n_fold: int) -> str:
        candidates = []
        for _ in range(n_fold):
            candidate = Candidate(
                self._engine,
                train_patterns,
                val_patterns,
                Queue())
            candidates.append(candidate)
        return Candidate.get_best(candidates)

    def _infer_with_cross_val_patterns(
            self, train_patterns: List[str], n_fold: int, total_train_rate: float) -> str:
        selected_train_count = int(len(train_patterns) * total_train_rate)
        train_buckets = Inference._get_train_buckets(
            train_patterns, selected_train_count, n_fold)
        candidates = []
        for i in range(n_fold):
            val_bucket = list(set(train_patterns) - set(train_buckets[i]))
            candidate = Candidate(
                self._engine,
                train_buckets[i],
                val_bucket,
                Queue())
            candidates.append(candidate)
        return Candidate.get_best(candidates)

    @staticmethod
    def _get_train_buckets(
            train_patterns: List[str], bucket_size: int, n_fold: int) -> List[List[str]]:
        if bucket_size <= len(train_patterns):
            train_selected = random.sample(
                train_patterns, bucket_size)
        else:
            train_selected = random.choices(
                train_patterns, k=bucket_size)
        train_buckets = list(
            chunked(
                train_selected,
                bucket_size //
                n_fold))
        return train_buckets



"""
TODO:
- [X] Evaluation Metrics Design
    - [X] F1-Score / Precision / Recall should work on whole Regex
- [X] Add LLMChain to fix the regex with low F1 scores (done by iterative calling).
- [ ] Consider continual inferencing mode: statistics should evaluate on the future cases.
"""
import typing
from typing import List, Optional, Dict
import re
from .filter import Filter
from .chain import Chain
from ..utils import make_verbose


class Engine:
    def __init__(self, openai_api_key: Optional[str] = None, temperature: float = 0.8,
                 mismatch_tolerance: float = 0.1, max_iteration: int = 3, simpify_regex: bool = True, verbose: bool = False):
        self._chain = Chain(
            openai_api_key=openai_api_key,
            temperature=temperature)
        self._mismatch_tolerance = mismatch_tolerance
        self._max_iteration = max_iteration
        self._simpify_regex = simpify_regex
        if verbose:
            self._make_verbose()
        self._verbose = verbose

    @typing.no_type_check
    def _make_verbose(self):
        self.run = make_verbose(self.run)
        self._run_new_inference = make_verbose(self._run_new_inference)
        self._fix_regex = make_verbose(self._fix_regex)

    def run(self, patterns: List[str]) -> str:
        regex_list = self.get_regex_sequence(patterns)
        return Engine.merge_regex_sequence(regex_list)

    def get_regex_sequence(self, patterns: List[str]) -> List[str]:
        assert len(
            patterns) > 0, '`patterns` input to `run` should no be an empty list'
        regex_list = [self._run_new_inference(patterns)]
        mismatched_patterns = Filter.mismatch(
            Engine.merge_regex_sequence(regex_list),
            patterns
        )
        while mismatched_patterns:
            regex = self._run_new_inference(mismatched_patterns)
            regex_list.append(regex)
            mismatched_patterns = Filter.mismatch(
                Engine.merge_regex_sequence(regex_list), patterns)
        return regex_list

    def _run_new_inference(self, patterns: List[str]) -> str:
        for _ in range(self._max_iteration):
            result = self._chain.inference_regex.run(
                Engine._convert_patterns_to_prompt(patterns)
            ).strip()
            try:
                re.compile(result)
                return result
            except KeyboardInterrupt as e:
                raise e
            except BaseException:
                pass
        raise ValueError(
            f'Unable to find inferred regex after {self._max_iteration} tries.')

    @staticmethod
    def _convert_patterns_to_prompt(patterns: List[str]) -> str:
        return '\n'.join(map(lambda x: f'"{x}"', patterns))

    @staticmethod
    def merge_regex_sequence(regex_list: List[str]) -> str:
        return '|'.join(map(lambda x: f'({x})', regex_list))

    def explain(self, regex: str) -> None:
        result = self._chain.explain_regex.run(regex)
        print(result)

    @staticmethod
    def get_correction_data(
            regex_list: List[str], patterns: List[str]) -> Dict[str, Dict[str, List[str]]]:
        """
        Args:
            - regex_list: the inference list of regex
            - the target patterns

        Returns:
            - correction_data (dict)
                - key: regex
                - value (dict)
                    - fields:
                        - correct
                        - incorrect
        """
        divided_patterns = Engine.divide_patterns(regex_list, patterns)
        result = dict()
        for i, regex in enumerate(regex_list):
            matched_patterns = Filter.match(regex_list[i], patterns)
            correct_patterns = divided_patterns[i]
            incorrect_patterns = list(
                set(matched_patterns) -
                set(correct_patterns))
            result[regex] = {
                'correct': correct_patterns,
                'incorrect': incorrect_patterns
            }
        return result

    @staticmethod
    def divide_patterns(regex_list: List[str],
                        patterns: List[str]) -> List[List[str]]:
        """
        Seperate a list of patterns to match the regex in regex_list
        """
        results = []
        for regex in regex_list:
            results.append(Filter.match(regex, patterns))
            patterns = Filter.mismatch(regex, patterns)
        return results

    def fix_regex_list(
            self, regex_list: List[str], correction_data: Dict[str, Dict[str, List[str]]]) -> List[str]:
        for i, regex in enumerate(regex_list):
            regex_list[i] = self.fix_regex(regex, correction_data)
        return regex_list

    def fix_regex(self, regex: str,
                  correction_data: Dict[str, Dict[str, List[str]]]) -> str:
        for _ in range(self._max_iteration):
            try:
                result = self._fix_regex(regex, correction_data)
                re.compile(result)
                break
            except KeyboardInterrupt as e:
                raise e
            except (ValueError, AssertionError):
                pass
        return result

    def _fix_regex(self, regex: str,
                   correction_data: Dict[str, Dict[str, List[str]]]) -> str:
        """
        Args:
            - regex_list: a list of regex to be fixed
            - correction_data: output of `get_correction_data`
        Return
            - fixed_regex_list: the corrected regex
        """
        regex_list = [regex]
        correct_patterns = [correction_data[regex]['correct']
                            for regex in regex_list]
        incorrect_patterns = [correction_data[regex]
                              ['incorrect'] for regex in regex_list]
        cnt = len(regex_list)
        fact_0_str = f"""
Fact 0:

A list of regex describing {cnt} type of patterns is double quoted and shown as the following bullet points:
    """
        regex_list_str = "\n".join(
            map(lambda x: f'{x[0]+1}. "{x[1]}"', enumerate(regex_list)))

        facts = "\n\n".join(map(lambda i: f"""
Fact {i+1}

For regex number {i+1}, it correctly match the patterns double quoted and shown as follows:

{Engine._convert_patterns_to_prompt(correct_patterns[i])}

However, it mistakenly match the patterns double quoted and shown as follows:

{Engine._convert_patterns_to_prompt(incorrect_patterns[i])}

""", range(cnt)))
        ans = self._chain.fix_regex.run(
            facts=f"""
{fact_0_str}

{regex_list_str}

Now, I will provide to you the other {cnt} facts.

{facts}
        """
        )
        if ans.endswith('""'):
            ans = ans[:-1]
        try:
            parsed_result = list(map(eval, ans.strip().split('\n')))
        except SyntaxError as e:
            raise ValueError(ans) from e

        assert len(regex_list) == len(parsed_result)
        for regex, result in zip(regex_list, parsed_result):
            try:
                assert regex == result[0], f'original regex is changed: {regex[0]}!={regex}'
                assert re.compile(result[1]), f'{result[1]} cannot be compiled'
            except BaseException as e:
                raise ValueError(f'Parsing result {result} failed') from e
        try:
            result = list(map(lambda x: x[1], parsed_result))
        except IndexError as e:
            raise ValueError(parsed_result) from e
        return result[0]



"""
How to sent this package onto PyPi?

1) Building Package Release Tar:
```python setup.py sdist```

2) Upload Package to PyPi:
```pip install twine```
```twine upload dist/*```
"""
import pathlib
import setuptools
from setuptools import find_packages

with open("README.md", "r") as fh:

    long_description = fh.read()


with open(str(pathlib.Path(__file__).parent.absolute()) +
          "/regex_inference/version.py", "r") as fh:
    version = fh.read().split("=")[1].replace("'", "")

setuptools.setup(

    name="regex-inference",

    version=version,

    author="jeffreylin",

    author_email="jeffrey82221@gmail.com",

    description="Regex Inference Engine based on ChatGPT",

    long_description=long_description,

    long_description_content_type="text/markdown",

    url="https://github.com/jeffrey82221/regex_inference",

    packages=find_packages(exclude=('tests',)),

    classifiers=[

        "Programming Language :: Python :: 3",

        "License :: OSI Approved :: MIT License",

        "Operating System :: OS Independent",

    ],
    python_requires='>=3.8, !=3.11.*',
    tests_require=['pytest'],
    install_requires=[
        'openapi-schema-pydantic==1.2.4',
        'pydantic==1.10.12',
        'openai==0.27.8',
        'langchain==0.0.253',
        'random_regex',
        'RegexFactory==1.0.0',
        'fado==2.0.4',
        'more_itertools==8.14.0'
    ]
)



"""
Inference Engine Mixing FAdo automata tool and AI
"""
import re
from typing import List
from functools import reduce
from regexfactory import escape
from FAdo.reex import RegExp, CDisj, CAtom, CConcat, CEpsilon, COption
from FAdo.conversions import FA2regexpCG
from .engine import Engine


__all__ = ['FAdoEngine', 'FAdoAIEngine']


class FAdoEngine(Engine):
    """
    Engine that infer regex using solely FAdo.
    """

    def _run_new_inference(self, patterns: List[str]) -> str:
        regex = FAdoEngine.infer_by_fado(patterns)
        re.compile(regex)
        return regex

    @staticmethod
    def infer_by_fado(inputs: List[str]) -> str:
        union_regex = FAdoEngine._make_regex_union(inputs)
        minimal_dfa = union_regex.nfaPD().toDFA().minimal()
        fado_regex = FA2regexpCG(minimal_dfa)
        standard_regex = FAdoEngine._to_simplied_standard_regex(fado_regex)
        return standard_regex

    @staticmethod
    def _make_regex_union(inputs: List[str]) -> RegExp:
        fado_regex_list = map(FAdoEngine._convert_str_to_fado_regex, inputs)
        return reduce(lambda x, y: CDisj(x, y), fado_regex_list)

    @staticmethod
    def _convert_str_to_fado_regex(input_str: str) -> CConcat:
        if input_str == '':
            return CEpsilon()
        else:
            atoms = []
            for ch in input_str:
                atoms.append(CAtom(escape(ch).regex))
            return reduce(lambda x, y: CConcat(x, y), atoms)

    @staticmethod
    def _to_simplied_standard_regex(regex: RegExp) -> str:
        standard_regex = FAdoEngine._to_standard_regex(regex)
        standard_regex = reduce(lambda x, y: x.replace(
            *y), [standard_regex, *list(FAdoEngine._generate_digit_range())])
        return standard_regex

    @staticmethod
    def _to_standard_regex(regex: RegExp) -> str:
        if isinstance(regex, CAtom):
            return regex.val
        elif isinstance(regex, COption):
            content = regex.arg
            regex = FAdoEngine._to_standard_regex(content)
            if isinstance(content, CAtom):
                return f'{regex}?'
            elif isinstance(content, CDisj) and regex[0] == '[' and regex[-1] == ']':
                return f'{regex}?'
            else:
                return f'({regex})?'
        elif isinstance(regex, CConcat):
            return FAdoEngine._to_standard_regex(
                regex.arg1) + FAdoEngine._to_standard_regex(regex.arg2)
        elif isinstance(regex, CDisj):
            if isinstance(regex.arg1, CEpsilon):
                return FAdoEngine._to_standard_regex(COption(regex.arg2))
            if isinstance(regex.arg2, CEpsilon):
                return FAdoEngine._to_standard_regex(COption(regex.arg1))
            x1, x2 = FAdoEngine._to_standard_regex(
                regex.arg1), FAdoEngine._to_standard_regex(
                regex.arg2)
            if isinstance(regex.arg1, CAtom) and isinstance(regex.arg2, CAtom):
                set_str = f'{x1}{x2}'
                set_str = ''.join(sorted(set_str))
                return f'[{set_str}]'
            elif isinstance(regex.arg1, CDisj) and x1[0] == '[' and x1[-1] == ']' and isinstance(regex.arg2, CAtom):
                x1 = x1[1:-1]
                set_str = f'{x1}{x2}'
                set_str = ''.join(sorted(set_str))
                return f'[{set_str}]'
            elif isinstance(regex.arg1, CDisj) and x1[0] == '(' and x1[-1] == ')':
                x1 = x1[1:-1]
                return f'({x1}|{x2})'
            else:
                return f'({x1}|{x2})'

        elif isinstance(regex, CEpsilon):
            return '[]'
        else:
            raise TypeError(f'type of regex: {type(regex)} cannot be handled')

    @staticmethod
    def _generate_digit_range():
        for i in range(10):
            for j in range(i + 2, 10):
                content = ''.join([str(e) for e in range(i, j + 1)])
                if i == 0 and j == 9:
                    yield f'[{content}]', '\\d'
                else:
                    yield f'[{content}]', f'[{i}-{j}]'


class FAdoAIEngine(FAdoEngine):
    """
    Engine that infer regex using both FAdo and ChatGPT
    """

    def _run_new_inference(self, patterns: List[str]) -> str:
        regex = FAdoEngine.infer_by_fado(patterns)
        if self._verbose:
            print('[_run_new_inference] infered fado regex:', regex)
        for _ in range(self._max_iteration):
            result = self._run_simplify_regex(regex).strip()
            try:
                re.compile(result)
                return result
            except KeyboardInterrupt as e:
                raise e
            except BaseException:
                if self._verbose:
                    print('[_run_new_inference] regex not working:', result)
                pass
        raise ValueError(
            f'Unable to find inferred regex after {self._max_iteration} tries.')

    def _run_simplify_regex(self, regex: str) -> str:
        for _ in range(self._max_iteration):
            result = self._chain.simplify_regex.run(
                regex=regex
            ).strip()
            try:
                re.compile(result)
                return result
            except KeyboardInterrupt as e:
                raise e
            except BaseException:
                pass
        raise ValueError(
            f'Unable to find simplified regex after {self._max_iteration} tries.')
